# @author NoCreativeIdeaForGoodusername
# encoding: utf-8
# module deepof

"""Plotting utility functions for the deepof package."""
import calendar
import copy
import os
import re
import time
import warnings
import itertools
from pathlib import Path
from typing import Any, List, NewType, Tuple, Union, Optional, NamedTuple
import dataclasses
from dataclasses import dataclass, field
from tqdm import tqdm
import cv2
import matplotlib.pyplot as plt
from matplotlib.animation import FFMpegWriter, FuncAnimation
import numpy as np
import pandas as pd
import seaborn as sns
from IPython.display import clear_output
from matplotlib.patches import Ellipse
import deepof.visuals_utils
from natsort import os_sorted

import deepof.post_hoc
import deepof.utils
from deepof.data_loading import get_dt
from deepof.config import PROGRESS_BAR_FIXED_WIDTH, BODYPART_COLORS
from deepof.visuals_utils import hex_to_BGR, BGR_to_hex, RGB_to_hex



# DEFINE CUSTOM ANNOTATED TYPES #
project = NewType("deepof_project", Any)
coordinates = NewType("deepof_coordinates", Any)
table_dict = NewType("deepof_table_dict", Any)


@dataclass
class VideoExportConfig:
    """Configuration for video annotations."""
    display_behavior_names: bool = True
    display_video_name: bool = False
    display_time: bool = False
    display_counter: bool = False
    display_arena: bool = False
    display_markers: bool = False
    display_mouse_labels: bool = False
    display_loading_bar: bool = True
    supervised_export: bool = True


@dataclass
class VideoExportProps:
    """Parameters for drawing text and shapes on the video frame."""
    font: int = cv2.FONT_HERSHEY_DUPLEX
    font_scale: float = 0.5
    thickness: int = 1
    padding: int = 5
    text_color: Tuple[int, int, int] = (255, 255, 255)
    outline_color: Tuple[int, int, int] = (0, 0, 0)
    arena_color: Tuple[int, int, int] = (40, 86, 236)
    arena_thickness: int = 3
    marker_radius: int = 3


def _filter_videos_by_condition(
    coordinates: coordinates,
    experiment_ids: List[str],
    conditions: dict[str, Any],
) -> List[str]:
    """Return a list of experiment IDs that match provided conditions."""
    if not conditions:
        return experiment_ids
    all_exp_conditions = coordinates.get_exp_conditions
    assert isinstance(conditions,dict), "Error! To select one experiment condition for export you must enter a dictionary {condition name: experiment condition [optional:, second condition name: second experiment condition, ...]}"
    list(conditions.values())
    filtered_conditions = [
        exp_id for exp_id in experiment_ids if
        all([all_exp_conditions.get(exp_id, {})[cond][0] == state
        for cond, state in conditions.items()])
    ]
    assert len(filtered_conditions)>0, f"No conditions are left after filtering. Make sure that your dictionary keys are among: {list(all_exp_conditions[list(all_exp_conditions.keys())[0]].columns)} and condition values are spelled correctly!" 
    return filtered_conditions

def _determine_behaviors_to_process(
    behavior_dict: dict,
    behaviors: Union[str, List[str]],
    behaviors_renamed: Union[str, List[str]],
) -> List[str]:
    """Determine the final lists of available, selected and renamed behaviors to generate videos for."""
    if isinstance(behaviors, str):
        return [behaviors]
    meta_info = get_dt(behavior_dict, list(behavior_dict.keys())[0], only_metainfo=True)
    if meta_info.get('columns') is None:
        available_behaviors = ["Cluster_"+ str(k) for k in range(meta_info.get('num_cols'))]
    else:
        available_behaviors = meta_info.get('columns')
    if behaviors is None:
        return available_behaviors, available_behaviors, available_behaviors
    return (
        available_behaviors, 
        [b for b in behaviors if b in available_behaviors], # all behaviors that occur in available_behaviors
        [b for i, b in enumerate(behaviors_renamed) if behaviors[i] in available_behaviors] # all renamed_behaviors that correspond to behaviors occuring in available_behaviors
    )


def output_videos_per_cluster(
    coordinates: coordinates,
    exp_conditions: dict,
    behavior_dict: dict,
    behaviors: Union[str, list],
    behaviors_renamed: list,
    frame_limit_per_video: int = float('inf'),
    bin_info: dict = None,
    roi_number: int = None,
    animals_in_roi: list = None,
    single_output_resolution: tuple = None,
    min_confidence: float = 0.0,
    min_bout_duration: int = None,
    video_export_config: VideoExportConfig = VideoExportConfig(),
    out_path: str = ".",
    roi_mode: str = "mousewise",
):
    """
    Generates one consolidated video per behavior, compiled from multiple experiments.
    """
    # Define the manual progress bar locally as it's a specific workaround
    def _loading_basic(current: int, total: int, bar_length: int = 68):
        if total == 0: return
        progress = (current + 1) / total
        filled_length = int(bar_length * progress)
        arrow = '>' if filled_length < bar_length else ''
        bar = f"[{'=' * filled_length}{arrow}{' ' * (bar_length - filled_length - len(arrow))}]"
        percent = f' {progress:.0%}'
        print(bar + percent, end='\r')
        if current == total - 1:
            print()

    output_path = Path(out_path)
    output_path.mkdir(parents=True, exist_ok=True)

    video_paths = coordinates.get_videos(full_paths=True)
    frame_rate = coordinates._frame_rate

    exp_ids_to_process = _filter_videos_by_condition(
        coordinates, list(behavior_dict.keys()), exp_conditions
    )
    available_behaviors, behaviors_to_process, behaviors_renamed_to_process  = _determine_behaviors_to_process(behavior_dict, behaviors, behaviors_renamed)
    
    video_export_config = dataclasses.replace(
        video_export_config, display_behavior_names=False, display_counter=False,
        display_video_name=True, display_loading_bar=False,
    )

    for behavior in tqdm(behaviors_to_process, desc=f"{'Exporting behavior videos':<{PROGRESS_BAR_FIXED_WIDTH}}", unit=" video"):

        behavior_renamed = behaviors_renamed_to_process[behaviors_to_process.index(behavior)]
        
        video_out_path = output_path / f"Behavior={behavior_renamed}_threshold={min_confidence}_{int(time.time())}.mp4"
        out = cv2.VideoWriter(
            str(video_out_path), cv2.VideoWriter_fourcc(*"mp4v"),
            frame_rate, single_output_resolution,
        )

        try:
            print("") # Newline to separate tqdm from manual bar
            
            total_exps = len(exp_ids_to_process)
            for i, exp_id in enumerate(exp_ids_to_process):
                _loading_basic(i, total_exps)
                
                cur_tab = get_dt(behavior_dict, exp_id)
                # Data frame conversion as soft counts get safed as numpy arrays
                if type(cur_tab)==np.ndarray:
                    cur_tab=pd.DataFrame(cur_tab,columns=available_behaviors)  

                # Behavior mask will function as behavior dataframe after float conversion
                behavior_mask, confidence = deepof.utils.get_behavior_mask_and_confidence(
                    cur_tab, behavior, video_export_config.supervised_export
                )
                behavior_mask_np=np.array(behavior_mask).squeeze()
                confidence=np.array(confidence).squeeze()

                confidence_indices = np.ones(len(behavior_mask), dtype=bool)
                confidence_indices = deepof.utils.filter_short_bouts(
                   behavior_mask_np, 
                    confidence, 
                    confidence_indices,
                    min_confidence, 
                    min_bout_duration
                )
                
                frames_passing_confidence = np.where(behavior_mask_np & confidence_indices)[0]

                if bin_info is not None:
                    
                    if roi_number is not None:
                        behavior_for_roi = behavior if roi_mode == "behaviorwise" else None
                        frames_in_roi = deepof.visuals_utils.get_behavior_frames_in_roi(
                            behavior=behavior_for_roi, local_bin_info=bin_info[exp_id],
                            animal_ids=animals_in_roi,
                        )
                        selected_frames = np.intersect1d(
                            frames_passing_confidence, frames_in_roi, assume_unique=True
                        )
                    else:
                        selected_frames = np.intersect1d(
                            frames_passing_confidence, bin_info[exp_id]["time"], assume_unique=True
                        )

                else:
                    selected_frames = frames_passing_confidence
                
                if len(selected_frames) > 0:
                    cap = cv2.VideoCapture(video_paths[exp_id])
                    output_annotated_video(
                        coordinates=coordinates, experiment_id=exp_id, tab=behavior_mask.astype(float),
                        behaviors=[behavior], video_export_config=video_export_config,
                        frames=selected_frames, cap=cap, out=out,
                        v_width=single_output_resolution[0], v_height=single_output_resolution[1],
                        frame_limit=frame_limit_per_video, out_path=output_path,
                        behaviors_renamed=[behavior_renamed],
                    )
        finally:
            out.release()
        
        clear_output()
    
    if hasattr(deepof.visuals_utils.get_behavior_frames_in_roi, '_warning_issued'):
        deepof.visuals_utils.get_behavior_frames_in_roi._warning_issued = False


def output_videos_per_cluster_old(
    coordinates: coordinates,
    exp_conditions: dict,
    behavior_dict: table_dict,
    behaviors: Union[str,list],
    behavior_names: list,
    frame_limit_per_video: int = np.inf,
    bin_info: dict = None,
    roi_number: int = None,
    animals_in_roi: list = None,
    single_output_resolution: tuple = None,
    min_confidence: float = 0.0,
    min_bout_duration: int = None,
    config: VideoExportConfig = VideoExportConfig(),
    out_path: str = ".",
    roi_mode: str = "mousewise",
): # pragma: no cover
    """Given a list of videos, and a list of soft counts per video, outputs a video for each cluster.

    Args:
        coordinates (coordinates): coordinates object for the current project. Used to get video paths.
        exp_conditions (dict): if provided, data coming from a particular condition is used. If not, all conditions are exported. If a dictionary with more than one entry is provided, the intersection of all conditions (i.e. male, stressed) is used.
        behavior_dict: table_dict containing data tables with behavior information (presence or absence of behaviors (columns) for each frame (rows))
        behaviors (Union[str,list]): list of behaviors to annotate
        behavior_names (list): Names of behaviors, potentially renamed by user.
        frame_limit_per_video: number of frames to render per video.
        bin_info (dict): dictionary containing indices to plot for all experiments
        roi_number (int): Number of the ROI that should be used for the plot (all behavior that occurs outside of the ROI gets excluded) 
        animals_in_roi (list): List of ids of the animals that need to be inside of the active ROI. All frames in which any of the given animals are not inside of the ROI get excluded 
        single_output_resolution: if single_output is provided, this is the resolution of the output video.
        min_confidence: minimum confidence threshold for a frame to be considered part of a cluster.
        min_bout_duration: minimum duration of a bout to be considered.
        display_time (bool): Displays current time in top left corner of the video frame
        display_arena (bool): Displays arena for each video.
        display_markers (bool): Displays mouse body parts on top of the mice. 
        display_mouse_labels (bool): Displays identities of the mice
        out_path: path to the output directory.
        roi_mode (str): Determines how the rois should be applied to different behaviors. Options are "mousewise" (default, selected mice needs to be inside the ROI) and "behaviorwise" (only mice involved in a behavior need to be inside of the ROI, only for supervised behaviors)                
    """

    #manual laoding bar for inner loop
    def _loading_basic(current, total, bar_length=68):
        progress = (current + 1) / total
        filled_length = int(bar_length * progress)
        arrow = '>' if filled_length < bar_length else ''
        bar = '[' + '=' * filled_length + arrow + ' ' * (bar_length - filled_length - len(arrow)) + ']'
        percent = f' {progress:.0%}'
        print(bar + percent, end='\r')
        if current == total - 1:
            print()  # Newline when complete

    def filter_experimental_conditions(
        coordinates: coordinates, videos: list, conditions: list
    ):
        """Return a list of videos that match the provided experimental conditions."""
        filtered_videos = videos

        for condition, state in conditions.items():

            filtered_videos = [
                video
                for video in filtered_videos
                if state
                == np.array(
                    coordinates.get_exp_conditions[re.findall("(.+)DLC", video)[0]][
                        condition
                    ]
                )
            ]

        return filtered_videos

    video_paths = filter_experimental_conditions(
        coordinates, coordinates.get_videos(full_paths=True), exp_conditions
    )
    frame_rate=coordinates._frame_rate

    meta_info=get_dt(behavior_dict,list(behavior_dict.keys())[0],only_metainfo=True)
    if isinstance(behaviors, str):
        behaviors = [behaviors]
    elif meta_info.get('columns') is not None and behaviors is None:
        behaviors = meta_info['columns']
    elif meta_info.get('columns') is not None:
        behaviors =[behavior for behavior in behaviors if behavior in meta_info['columns']]
    elif behaviors is not None:
        behaviors = [behavior for behavior in behaviors if behavior in behavior_names] 
    else:
        behaviors=behavior_names

    # Iterate over all clusters, and output a masked video for each
    for cur_behavior in tqdm(behaviors, desc=f"{'Exporting behavior videos':<{PROGRESS_BAR_FIXED_WIDTH}}", unit="video"):
    
        #creates a new line to ensure that the outer loading bar does not get overwritten by the inner one
        print("")

        out = cv2.VideoWriter(
            os.path.join(
                out_path,
                "Behavior={}_threshold={}_{}.mp4".format(
                    cur_behavior, min_confidence, calendar.timegm(time.gmtime())
                ),
            ),
            cv2.VideoWriter_fourcc(*"mp4v"),
            frame_rate,
            single_output_resolution,
        )
      
        bar_len=len(behavior_dict.keys())
        #this loop uses a manual loading bar as tqdm does not work here for some reason
        for i, key in enumerate(behavior_dict.keys()):

            _loading_basic(i, bar_len)
            
            cur_tab=copy.deepcopy(get_dt(behavior_dict, key))

            # Turn numpy arrays (in case of unsupervised data) into dataframes
            if type(cur_tab)==np.ndarray:
                cur_tab=pd.DataFrame(cur_tab,columns=behavior_names) 

                # Get positions at which the current cluster is the most likely one 
                max_entry_columns = cur_tab.idxmax(axis=1)
                behavior_mask = (cur_behavior == max_entry_columns)
                idx = (cur_behavior == max_entry_columns)
                confidence = cur_tab[cur_behavior]    
        
            else:
                cur_tab.columns = behavior_names

                # Get positions at which the current behaviro occurs
                behavior_mask = cur_tab[cur_behavior]>0.1
                idx = cur_tab[cur_behavior]>0.1
                confidence = cur_tab[cur_behavior]  

            # Convert mask to contain either selected behavior or nothing for each frame     
            behavior_mask = behavior_mask.astype(str)
            behavior_mask[idx]=str(cur_behavior)
            behavior_mask[~idx]=""
          
            # Get hard counts and confidence estimates per cluster
            confidence_indices = np.ones(behavior_mask.shape[0], dtype=bool)

            # Given a frame mask, output a subset of the given video to disk, corresponding to a particular cluster
            cap = cv2.VideoCapture(video_paths[key])
            v_width, v_height = single_output_resolution

            # Compute confidence mask, filtering out also bouts that are too short
            confidence_indices = deepof.utils.filter_short_bouts(
                idx.astype(float),
                confidence,
                confidence_indices,
                min_confidence,
                min_bout_duration,
            )
            confidence_mask = (behavior_mask == str(cur_behavior)) & confidence_indices

            # get frames for current video
            frames = None
            if bin_info is not None:
                if roi_number is not None:
                    if roi_mode == "behaviorwise":
                        behavior_in=behaviors[0]
                    else:
                        behavior_in=None
                    frames=deepof.visuals_utils.get_behavior_frames_in_roi(behavior=behavior_in, local_bin_info=bin_info[key], animal_ids=animals_in_roi)
                else:
                    frames=bin_info[key]["time"]          

            selected_frames=frames[confidence_mask[frames]]
            
            # Sdjust config for cluster export
            config.display_behavior_names=False
            config.display_counter=False
            config.display_video_name=True
            config.display_loading_bar=False

            if len(selected_frames)>0:
                output_annotated_video(
                    coordinates=coordinates,
                    experiment_id=key,                
                    tab=cur_tab,
                    behaviors=[cur_behavior],
                    video_export_config=config,
                    frames=selected_frames,
                    cap=cap,
                    out=out,
                    v_width=v_width,
                    v_height=v_height,
                    frame_limit=frame_limit_per_video,
                    out_path=out_path,
                )
        

        out.release()
        #to not flood the output with loading bars
        clear_output()
    deepof.visuals_utils.get_behavior_frames_in_roi._warning_issued = False


def _prepare_behavior_dataframe(
    tab: pd.DataFrame, behaviors: List[str], behavior_renamed: List[str],
) -> pd.DataFrame:
    """
    Creates a DataFrame where cells contain the behavior name if the condition
    is met (score > 0.1), and an empty string otherwise. This is an efficient
    way to look up the active behavior string for a given frame.
    """
    behavior_df = pd.DataFrame(index=tab.index, columns=behaviors, data="")
    for behavior, behavior_renamed in zip(behaviors, behavior_renamed):
        mask = tab[behavior] > 0.1
        behavior_df.loc[mask, behavior] = behavior_renamed
    return behavior_df


def _draw_arena(
    frame: np.ndarray,
    arena_type: str,
    arena_params: Any,
    params: VideoExportProps
):
    """Draws the arena boundaries on the frame."""
    if isinstance(arena_params, Tuple): #Circular (legacy)  arena_type.startswith("circular"):
        cv2.ellipse(
            img=frame,
            center=np.round(arena_params[0]).astype(int),
            axes=np.round(arena_params[1]).astype(int),
            angle=arena_params[2],
            startAngle=0,
            endAngle=360,
            color=params.arena_color,
            thickness=params.arena_thickness,
        )
    elif isinstance(arena_params, np.ndarray): #Polygonal
        cv2.polylines(
            img=frame,
            pts=[np.array(arena_params, dtype=np.int32)],
            isClosed=True,
            color=params.arena_color,
            thickness=params.arena_thickness,
        )


def _draw_markers(
    frame: np.ndarray,
    coords_at_frame: pd.Series,
    animal_ids: List[str],
    params: VideoExportProps
):
    """Draws body part markers on the frame."""
    for bpart in coords_at_frame.index.get_level_values(0).unique():
        if not np.isnan(coords_at_frame.loc[(bpart, "x")]):
            # Determine color by finding which animal this body part belongs to
            color_idx = [bpart.startswith(id) for id in animal_ids].index(True)
            color = BODYPART_COLORS[color_idx]
            cv2.circle(
                frame,
                (int(coords_at_frame.loc[(bpart, "x")]), int(coords_at_frame.loc[(bpart, "y")])),
                radius=params.marker_radius,
                color=color,
                thickness=-1,  # Filled circle
            )


def _draw_mouse_labels(
    frame: np.ndarray,
    coords_at_frame: pd.Series,
    animal_ids: List[str],
    params: VideoExportProps,
):
    """Draws labels with mouse identities on the frame."""
    for mouse_id in animal_ids:
        center_bpart = f"{mouse_id}_Center"
        if center_bpart in coords_at_frame.index and not np.isnan(coords_at_frame.loc[(center_bpart, "x")]):
            mouse_pos = (
                int(coords_at_frame.loc[(center_bpart, "x")]),
                int(coords_at_frame.loc[(center_bpart, "y")])
            )
            (text_w, text_h), baseline = cv2.getTextSize(
                mouse_id, params.font, params.font_scale, params.thickness
            )
            color_idx = animal_ids.index(mouse_id)
            color = BODYPART_COLORS[color_idx]

            top_left = (mouse_pos[0], mouse_pos[1] - text_h - params.padding)
            bottom_right = (mouse_pos[0] + text_w, mouse_pos[1] + baseline)
            
            cv2.rectangle(frame, top_left, bottom_right, color, -1)
            cv2.putText(
                frame, mouse_id, mouse_pos, params.font,
                params.font_scale, params.text_color, params.thickness
            )


def _draw_behavior_info(
    frame: np.ndarray,
    frame_idx: int,
    v_width: int,
    behavior_df: pd.DataFrame,
    behaviors: List[str],
    behavior_colors: List[str],
    behavior_counters: np.ndarray,
    widest_text_size: Tuple[int, int],
    shift_name_box: bool,
    config: VideoExportConfig,
    params: VideoExportProps,
    frame_rate: int
):
    """Draws the behavior names, background boxes, and counters."""
    text_width, text_height = widest_text_size
    y_start = 10 + text_height
    line_step = int(text_height * 2)

    for i, behavior in enumerate(behaviors):
        behavior_text = behavior_df.iloc[frame_idx][i]
        if behavior_text:
            # Fixed vertical position per behavior when shifting is enabled
            box_y = y_start + (i * line_step if shift_name_box else 0)

            # Draw background rectangle
            top_left = (v_width - text_width - params.padding * 2, box_y - text_height - params.padding)
            bottom_right = (v_width - params.padding, box_y + params.padding)
            cv2.rectangle(frame, top_left, bottom_right, hex_to_BGR(behavior_colors[i]), -1)

            # Update and format text with counter if enabled
            if config.display_counter:
                behavior_counters[i] += 1
                time_str = deepof.visuals_utils.seconds_to_time(
                    behavior_counters[i] / frame_rate, cut_milliseconds=False
                )[3:11]
                behavior_text += f' {time_str}'

            # Draw behavior text
            text_pos = (v_width - text_width - params.padding, box_y)
            cv2.putText(frame, behavior_text, text_pos, params.font, params.font_scale, params.text_color, params.thickness)


def output_annotated_video(
    coordinates: coordinates,
    experiment_id: str,
    tab: pd.DataFrame,
    behaviors: List[str],
    video_export_config: VideoExportConfig = VideoExportConfig(),
    frames: np.array = None,
    cap: Any = None,
    out: Any = None,
    v_width: int = None,
    v_height: int = None,
    frame_limit: int = float('inf'),
    out_path: Path = Path("."),
    behaviors_renamed: List = None,
):
    """
    Generates a video with frames annotated with specified behaviors and other metadata.

    Args:
        coordinates: Coordinates object for the project, used to access video paths and metadata.
        experiment_id: ID of the experiment to export.
        tab: DataFrame with behavior probabilities/scores per frame.
        behaviors: A list of behavior names (columns in `tab`) to annotate.
        video_export_config: A dataclass object specifying video export information (what to display, export mode).
        frames: An array of specific frame indices to include in the output video.
        cap: An existing cv2.VideoCapture object. If None, one will be created.
        out: An existing cv2.VideoWriter object. If None, one will be created.
        v_width: Desired output video width. Defaults to source video width.
        v_height: Desired output video height. Defaults to source video height.
        frame_limit: Maximum number of frames to process.
        out_path: The directory where the output video will be saved.
        behaviors_renamed: List of updated behavior names for display
    """
    video_path = Path(coordinates.get_videos(full_paths=True)[experiment_id])
    video_name_stem = video_path.stem

    # --- Behavior & Frame Preparation ---
    shift_name_box = True
    if behaviors is None:
        # If no behaviors are specified, check if it's a single-behavior-per-frame scenario
        if not (tab.sum(axis=1) > 1.9).any():
            behaviors = list(tab.columns)
            shift_name_box = False  # Display all behaviors simultaneously
        else:
            behaviors = list(tab.columns)
            #raise ValueError("A list of 'behaviors' must be provided for multi-label annotations.")

    if behaviors_renamed is None or len(behaviors_renamed) != len(behaviors):
        behaviors_renamed=behaviors

    # Rename behaviors to user given names 
    behavior_df = _prepare_behavior_dataframe(tab=tab, behaviors=behaviors,behavior_renamed=behaviors_renamed)
    
    cur_coords = get_dt(coordinates._tables, experiment_id)

    # Filter frames to ensure they are within bounds and respect the frame limit
    if frames is not None:
        frames = frames[frames < len(behavior_df)]
        if len(frames) > frame_limit:
            frames = frames[:frame_limit]
    else: # If no frames are given, process all frames up to the limit
        total_frames = min(len(behavior_df), frame_limit)
        frames = np.arange(total_frames)

    # --- Video I/O Setup ---
    cap_is_external = cap is not None
    out_is_external = out is not None
    
    if not cap_is_external:
        cap = cv2.VideoCapture(str(video_path))
    
    # Determine video dimensions
    resize_frame = v_width is not None or v_height is not None
    if v_width is None:
        v_width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    if v_height is None:
        v_height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
    
    frame_rate = coordinates._frame_rate
    if not out_is_external:
        video_out_path = os.path.join(out_path, f"{video_name_stem}_annotated_{int(time.time())}.mp4")
        out = cv2.VideoWriter(
            str(video_out_path), cv2.VideoWriter_fourcc(*"mp4v"), frame_rate, (v_width, v_height)
        )

    # --- Drawing & Annotation Parameter Setup ---
    params = VideoExportProps()
    bg_colors = deepof.visuals_utils.get_behavior_colors(behaviors, behavior_df)
    behavior_counters = np.zeros(len(behaviors))

    # Pre-calculate text size for layout purposes
    widest_text = max(behaviors_renamed, key=len)
    if video_export_config.display_counter:
        widest_text += ' 00:00.00'  # Add placeholder for counter time
    (text_w, text_h), _ = cv2.getTextSize(widest_text, params.font, params.font_scale, params.thickness)
    widest_text_size = (text_w, text_h)

    # Pre-calculate scaled coordinates and arena parameters if needed
    scaling_ratio = coordinates._scales[experiment_id][2] / coordinates._scales[experiment_id][3]
    if video_export_config.display_arena:
        arena_params = coordinates._arena_params[experiment_id]
        if isinstance(arena_params, np.ndarray): # "polygonal" in coordinates._arena:
            scaled_arena_params = np.array(arena_params) * scaling_ratio
        elif isinstance(arena_params, Tuple):
            scaled_arena_params = (
                tuple(np.array(arena_params[0]) * scaling_ratio),
                tuple(np.array(arena_params[1]) * scaling_ratio),
                arena_params[2],
            )
    if video_export_config.display_markers or video_export_config.display_mouse_labels:
        scaled_coords = cur_coords * scaling_ratio

    # --- Main Processing Loop ---
    try:
        frame_indices = tqdm(range(len(frames)), desc=f"{'Exporting behavior video':<{PROGRESS_BAR_FIXED_WIDTH}}", unit="Frame") \
                        if video_export_config.display_loading_bar else range(len(frames))

        for i in frame_indices:
            frame_idx = frames[i]
           
            # Efficiently seek frames only when necessary (not consecutive)
            if i == 0 or frames[i] - frames[i-1] != 1:
                cap.set(cv2.CAP_PROP_POS_FRAMES, frame_idx)
            
            ret, frame = cap.read()
            if not ret:
                print(f"Warning: Could not read frame {frame_idx}. Stopping.")
                break

            # Annotations drawn BEFORE resizing
            if video_export_config.display_arena:
                _draw_arena(frame, coordinates._arena, scaled_arena_params, params)
            
            if video_export_config.display_markers:
                _draw_markers(frame, scaled_coords.iloc[frame_idx], coordinates._animal_ids, params)
            
            if video_export_config.display_mouse_labels:
                _draw_mouse_labels(frame, scaled_coords.iloc[frame_idx], coordinates._animal_ids, params)

            # Resize frame if custom dimensions are provided
            if resize_frame:
                frame = cv2.resize(frame, (v_width, v_height))
            
            # Annotations drawn AFTER resizing
            if video_export_config.display_behavior_names:
                _draw_behavior_info(
                    frame, frame_idx, v_width, behavior_df, behaviors_renamed, bg_colors,
                    behavior_counters, widest_text_size, shift_name_box, video_export_config, params, frame_rate
                )
            
            if video_export_config.display_video_name:
                cv2.putText(frame, video_name_stem, (15, v_height - 15),
                            params.font, 0.75, params.text_color, 2)
            
            if video_export_config.display_time:
                time_text = f"time: {deepof.visuals_utils.seconds_to_time(frame_idx / frame_rate)}"
                pos = (params.padding, 10 + text_h)
                cv2.putText(frame, time_text, pos, params.font, params.font_scale * 1.5, params.outline_color, params.thickness + 2)
                cv2.putText(frame, time_text, pos, params.font, params.font_scale * 1.5, params.text_color, params.thickness)

            out.write(frame)

    finally:
        # Ensure all resources are released
        if not cap_is_external:
            cap.release()
        if not out_is_external:
            out.release()
        cv2.destroyAllWindows()


def output_annotated_video_old(
    coordinates: coordinates,
    experiment_id: str,
    tab: np.ndarray,
    behaviors: list,
    frames: np.array = None,
    display_behavior_names: bool = True,
    display_video_name: bool = False,
    display_time: bool = False,
    display_counter: bool = False,
    display_arena: bool = False,
    display_markers: bool = False,
    display_mouse_labels: bool = False,
    display_loading_bar: bool = True,
    cap: Any = None,
    out: Any = None,
    v_width: int = None,
    v_height: int = None,
    frame_limit: int = np.inf,
    out_path: str = ".",
): # pragma: no cover
    """Given a video, and soft_counts per frame, outputs a video with the frames annotated with the cluster they belong to.

    Args:
        coordinates (coordinates): coordinates object for the current project. Used to get video paths.
        experiment_id: id of the experiment to export
        soft_counts: soft cluster assignments for a specific video
        behavior (str): Behavior or Cluster to that gets exported. If none is given, all Clusters get exported for softcounts and only nose2nose gets exported for supervised annotations.
        frames: frames that should be exported.
        display_behavior_names (bool): Display the names of the respective behaviors
        display_video_name (bool): Display the name of the video
        display_time (bool): Displays current time in top left corner of the video frame
        display_counter (bool): Displays event counter for each displayed event.   
        display_arena (bool): Displays arena for each video.
        display_markers (bool): Displays mouse body parts on top of the mice.
        display_mouse_labels (bool): Displays identities of the mice
        display_loading_bar (bool): Displays the laoding bar during writing of the video
        cap (Any): video capture object for reading the video, can be provided. Will be created from video at experiment_id otherwise.
        out (Any): video capture object for writing the video, can be provided.
        v_width (int): video width
        v_height (int): video height    
        frame_limit (int): Maximum number of frames that can be included in a video. No limit per default
        out_path: out_path: path to the output directory.

    """

    video_path=coordinates.get_videos(full_paths=True)[experiment_id]
    # for display
    re_path = re.findall(r".+[/\\]([^/.]+?)(?=\.|DLC)", video_path)[0]

    # if every frame has only one distinct behavior assigned to it, plot all behaviors
    shift_name_box=True
    if behaviors is None and not (np.sum(tab, 1)>1.9).any():
        behaviors = list(tab.columns)
        shift_name_box=False
    elif behaviors is None:
        raise ValueError(
            "Cannot accept no behavior for supervised annotations!"
        )
    # create behavior_df that lists in which frames each behavior occurs
    behavior_df = tab[behaviors]>0.1
    idx = tab[behaviors]>0.1 #OK, I know this looks weird, but it is actually not a bug and works
    behavior_df = behavior_df.astype(str)
    behavior_df[idx]=behaviors
    behavior_df[~idx]=""
    # Else if every frame has only one distinct behavior assigned to it, annotate all behaviors

    cur_coords=get_dt(coordinates._tables,experiment_id)
    
    # Ensure that no frames are requested that are outside of the provided data
    if np.max(frames) >= len(behavior_df):
        frames = np.where(frames<len(behavior_df))[0]
    if len(frames) >= frame_limit:
        frames = frames[0:frame_limit]

    # Given a frame mask, output a subset of the given video to disk, corresponding to a particular cluster
    if cap is None:
        cap = cv2.VideoCapture(video_path)

    # Get width and height of current video
    resize_frame=True    
    if v_width is None and v_height is None:
        resize_frame=False    
    if v_width is None:
        v_width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    if v_height is None:  
        v_height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))  

    video_out = os.path.join(
        out_path,
        os.path.split(video_path)[-1].split(".")[0] 
        + "_annotated_{}.mp4".format(calendar.timegm(time.gmtime())),
    )

    frame_rate=coordinates._frame_rate
    if out is None:
        out = cv2.VideoWriter(
            video_out, cv2.VideoWriter_fourcc(*"mp4v"), frame_rate, (v_width, v_height)
        )

    # Prepare text
    font = cv2.FONT_HERSHEY_DUPLEX
    font_scale = 0.5
    thickness = 1
    text_widths=[cv2.getTextSize(behavior, font, font_scale, thickness)[0][0] for behavior in behaviors]
    widest_text=behaviors[np.argmax(text_widths)]
    if display_counter:
        behavior_array=np.zeros(len(behaviors))
        widest_text=widest_text+' 00:00.00'
    if display_arena:
        arena_params = coordinates._arena_params[experiment_id]
        # scale arena_params back o video res
        scaling_ratio = coordinates._scales[experiment_id][2]/coordinates._scales[experiment_id][3]
        if isinstance(arena_params, np.ndarray):
            arena_params=np.array(arena_params)*scaling_ratio
        elif isinstance(arena_params, Tuple):
            # scale from mm to original pixel resolution
            arena_params=(tuple(np.array(arena_params[0])*scaling_ratio),tuple(np.array(arena_params[1])*scaling_ratio),arena_params[2])
    if display_markers or display_mouse_labels:
        scaling_ratio = coordinates._scales[experiment_id][2]/coordinates._scales[experiment_id][3]
        cur_coords=cur_coords*scaling_ratio
    (text_width, text_height), baseline = cv2.getTextSize(widest_text, font, font_scale, thickness)
    (text_width_time, text_height_time), baseline = cv2.getTextSize("time: 00:00:00", font, font_scale, thickness)
    x = 10  # 10 pixels from left
    y = 10 + text_height_time  # 10 pixels from top (accounting for text height)
    padding = 5
    bg_color = deepof.visuals_utils.get_behavior_colors(behaviors, tab)

    diff_frames = np.diff(frames)
    for i in (tqdm(range(len(frames)), desc=f"{'Exporting behavior video':<{PROGRESS_BAR_FIXED_WIDTH}}", unit="Frame")) if display_loading_bar else range(len(frames)):

        if i == 0 or diff_frames[i-1] != 1:
            cap.set(cv2.CAP_PROP_POS_FRAMES, frames[i])
        ret, frame = cap.read()

        if ret == False or cap.isOpened() == False:
            break

        try:

            #####
            # drawn annotations before rescaling
            #####
            if display_arena:
                
                if coordinates._arena.startswith("circular"):
                    # Print arena for debugging
                    cv2.ellipse(
                        img=frame,
                        center=np.round(arena_params[0]).astype(int),
                        axes=np.round(arena_params[1]).astype(int),
                        angle=arena_params[2],
                        startAngle=0,
                        endAngle=360,
                        color=(40, 86, 236),
                        thickness=3,
                    )

                elif coordinates._arena.startswith("polygonal"):

                    # Draw polygon
                    cv2.polylines(
                        img=frame,
                        pts=[np.array(arena_params, dtype=np.int32)],
                        isClosed=True,
                        color=(40, 86, 236),
                        thickness=3,
                    )


            if display_markers:
                # Print body parts for debuging
                for bpart in cur_coords.columns.levels[0]:
                    pass
                    if not np.isnan(cur_coords[bpart]["x"][frames[i]]):
                        cv2.circle(
                            frame,
                            (int(cur_coords[bpart]["x"][frames[i]]), int(cur_coords[bpart]["y"][frames[i]])),
                            radius=3,
                            color=(
                                BODYPART_COLORS[[bpart.startswith(id) for id in coordinates._animal_ids].index(True)] #first index of animal id that fits
                            ),
                            thickness=-1,
                        )
                
            
            if display_mouse_labels:

                for bpart in cur_coords.columns.levels[0]:

                    if bpart.endswith("Center") and not np.isnan(cur_coords[bpart]["x"][frames[i]]):

                        mouse_id=[id for id in coordinates._animal_ids if bpart.startswith(id)][0]
                        mouse_pos=(int(cur_coords[bpart]["x"][frames[i]]), int(cur_coords[bpart]["y"][frames[i]]))
                        (id_text_width, id_text_height), id_baseline = cv2.getTextSize(mouse_id, font, font_scale, thickness)

                        cv2.rectangle(frame, 
                            (mouse_pos[0], mouse_pos[1] - id_text_height - padding),  # Top-left corner
                            (mouse_pos[0] + id_text_width, mouse_pos[1] + id_baseline),  # Bottom-right corner
                            (BODYPART_COLORS[[bpart.startswith(id) for id in coordinates._animal_ids].index(True)]),  # Blue color (BGR format)
                            -1)  # Filled rectangle
                        
                        cv2.putText(frame, mouse_id, mouse_pos, font, font_scale, (255, 255, 255), thickness)



            #####
            #resize frame if resolution is specified, needs to be done after drawn annotations but before written annotations
            #####
            if resize_frame:
                frame = cv2.resize(frame, [v_width, v_height])


            #####
            # written annotations after rescaling
            #####
            if display_behavior_names:
                
                ystep=0
                for z, behavior in enumerate(behaviors):
                    if len(behavior_df[behavior][frames[i]])>0:
                        cv2.rectangle(frame, 
                            (v_width - text_width - x , y - text_height - padding +ystep),  # Top-left corner
                            (v_width - padding, y + baseline +ystep),  # Bottom-right corner
                            hex_to_BGR(bg_color[z]),  # Blue color (BGR format)
                            -1)  # Filled rectangle

                        behavior_text=str(behavior_df[behavior][frames[i]])
                        if display_counter:
                            behavior_array[z]=behavior_array[z]+1
                            behavior_text = behavior_text + ' ' + deepof.visuals_utils.seconds_to_time(behavior_array[z]/frame_rate, cut_milliseconds=False)[3:11]

                        # Draw white main text
                        cv2.putText(frame, str(behavior_text), (v_width - text_width - padding, y+ystep), font, font_scale, (255, 255, 255), thickness)
                    if shift_name_box:
                        ystep=ystep+int(text_height*2)


            if display_video_name:
                cv2.putText(
                    frame,
                    re_path,
                    (int(v_width * 0.3 / 10), int(v_height / 1.05)),
                    cv2.FONT_HERSHEY_DUPLEX,
                    0.75,
                    (255, 255, 255),
                    2,
                )


            if display_time:

                disp_time = "time: " + deepof.visuals_utils.seconds_to_time(frames[i]/frame_rate)
                # Draw black outline
                cv2.putText(frame, disp_time, (x, y), font, font_scale*1.5, (0, 0, 0), thickness + 2)
                # Draw white main text
                cv2.putText(frame, disp_time, (x, y), font, font_scale*1.5, (255, 255, 255), thickness)

            out.write(frame)
        except IndexError:
            ret = False

    cap.release()
    cv2.destroyAllWindows()

    #writevideo = FFMpegWriter(fps=frame_rate)
    #animation.save(save, writer=writevideo)

    return None


# @author lucasmiranda42
# encoding: utf-8
# module deepof

"""Utility functions for both training autoencoder models in deepof.models and tuning hyperparameters with deepof.hypermodels."""

import os
from datetime import date, datetime
from typing import Any, List, NewType, Tuple, Union
import copy

from IPython.display import clear_output
import matplotlib.pyplot as plt
import psutil
import numpy as np
import tensorflow as tf
import tensorflow.keras.backend as K
import tensorflow_probability as tfp
import tqdm
from keras_tuner import BayesianOptimization, Hyperband, Objective
from spektral.layers import CensNetConv
from tensorboard.plugins.hparams import api as hp
from tensorflow.keras.initializers import he_uniform
from tensorflow.keras.layers import (
    GRU,
    Bidirectional,
    LayerNormalization,
    TimeDistributed,
)

from deepof.config import PROGRESS_BAR_FIXED_WIDTH
import deepof.data
import deepof.hypermodels
import deepof.models
import deepof.post_hoc
from deepof.data_loading import get_dt, save_dt


tfb = tfp.bijectors
tfd = tfp.distributions
tfpl = tfp.layers

# Ignore warning with no downstream effect
tf.get_logger().setLevel("ERROR")
tf.autograph.set_verbosity(0)

# DEFINE CUSTOM ANNOTATED TYPES #
project = NewType("deepof_project", Any)
coordinates = NewType("deepof_coordinates", Any)
table_dict = NewType("deepof_table_dict", Any)

### CONTRASTIVE LEARNING UTILITIES
def select_contrastive_loss(
    history,
    future,
    similarity,
    loss_fn="nce",
    temperature=0.1,
    tau=0.1,
    beta=0.1,
    elimination_topk=0.1,
):  # pragma: no cover
    """Select and applies the contrastive loss function to be used in the Contrastive embedding models.

    Args:
        history: Tensor of shape (batch_size, seq_len, embedding_dim).
        future: Tensor of shape (batch_size, seq_len, embedding_dim).
        similarity: Function that computes the similarity between two tensors.
        loss_fn: String indicating the loss function to be used.
        temperature: Float indicating the temperature to be used in the specified loss function.
        tau: Float indicating the tau value to be used if DCL or hard DLC are selected.
        beta: Float indicating the beta value to be used if hard DLC is selected.
        elimination_topk: Float indicating the top-k value to be used if FC is selected.

    """
    similarity_dict = {
        "cosine": _cosine_similarity,
        "dot": _dot_similarity,
        "euclidean": _euclidean_similarity,
        "edit": _edit_similarity,
    }
    similarity = similarity_dict[similarity]

    if loss_fn == "nce":
        loss, pos, neg = nce_loss(history, future, similarity, temperature)
    elif loss_fn == "dcl":
        loss, pos, neg = dcl_loss(
            history, future, similarity, temperature, debiased=True, tau_plus=tau
        )
    elif loss_fn == "fc":
        loss, pos, neg = fc_loss(
            history,
            future,
            similarity,
            temperature,
            elimination_topk=elimination_topk,
        )
    elif loss_fn == "hard_dcl":
        loss, pos, neg = hard_loss(
            history,
            future,
            similarity,
            temperature,
            beta=beta,
            debiased=True,
            tau_plus=tau,
        )

    # noinspection PyUnboundLocalVariable
    return loss, pos, neg


def _cosine_similarity(x, y):  # pragma: no cover
    """Compute the cosine similarity between two tensors."""
    v = tf.keras.losses.CosineSimilarity(
        axis=2, reduction=tf.keras.losses.Reduction.NONE
    )(tf.expand_dims(x, 1), tf.expand_dims(y, 0))
    return -v


def _dot_similarity(x, y):  # pragma: no cover
    """Compute the dot product between two tensors."""
    v = tf.tensordot(tf.expand_dims(x, 1), tf.expand_dims(tf.transpose(y), 0), axes=2)

    return v


def _euclidean_similarity(x, y):  # pragma: no cover
    """Compute the euclidean distance between two tensors."""
    x1 = tf.expand_dims(x, 1)
    y1 = tf.expand_dims(y, 0)
    d = tf.sqrt(tf.reduce_sum(tf.square(x1 - y1), axis=2))
    s = 1 / (1 + d)
    return s


def _edit_similarity(x, y):  # pragma: no cover
    """Compute the edit distance between two tensors."""
    x1 = tf.expand_dims(x, 1)
    y1 = tf.expand_dims(y, 0)
    d = tf.sqrt(tf.reduce_sum(tf.square(x1 - y1), axis=2))
    s = 1 / (1 + d)
    return s


def nce_loss(history, future, similarity, temperature=0.1):  # pragma: no cover
    """Compute the NCE loss function, as described in the paper "A Simple Framework for Contrastive Learning of Visual Representations" (https://arxiv.org/abs/2002.05709)."""
    criterion = tf.keras.losses.BinaryCrossentropy(
        from_logits=True, reduction=tf.keras.losses.Reduction.SUM
    )

    N = history.shape[0]
    sim = similarity(history, future)
    pos_sim = K.exp(tf.linalg.tensor_diag_part(sim) / temperature)

    tri_mask = np.ones(N**2, dtype=bool).reshape(N, N)
    tri_mask[np.diag_indices(N)] = False
    neg = tf.reshape(tf.boolean_mask(sim, tri_mask), [N, N - 1])
    all_sim = K.exp(sim / temperature)

    logits = tf.divide(K.sum(pos_sim), K.sum(all_sim, axis=1))

    lbl = np.ones(history.shape[0])

    # categorical cross entropy
    loss = criterion(y_pred=logits, y_true=lbl)

    # similarity of positive pairs (only for debug)
    mean_sim = K.mean(tf.linalg.tensor_diag_part(sim))
    mean_neg = K.mean(neg)
    return loss, mean_sim, mean_neg


def dcl_loss(
    history, future, similarity, temperature=0.1, debiased=True, tau_plus=0.1
):  # pragma: no cover
    """Compute the DCL loss function, as described in the paper "Debiased Contrastive Learning" (https://github.com/chingyaoc/DCL/)."""
    N = history.shape[0]
    sim = similarity(history, future)
    pos_sim = K.exp(tf.linalg.tensor_diag_part(sim) / temperature)

    tri_mask = np.ones(N**2, dtype=bool).reshape(N, N)
    tri_mask[np.diag_indices(N)] = False
    neg = tf.reshape(tf.boolean_mask(sim, tri_mask), [N, N - 1])
    neg_sim = K.exp(neg / temperature)

    # estimator g()
    if debiased:
        N = N - 1
        Ng = (-tau_plus * N * pos_sim + K.sum(neg_sim, axis=-1)) / (1 - tau_plus)
        # constraint (optional)
        Ng = tf.clip_by_value(
            Ng,
            clip_value_min=N * np.e ** (-1 / temperature),
            clip_value_max=tf.float32.max,
        )
    else:
        Ng = K.sum(neg_sim, axis=-1)

    # contrastive loss
    loss = K.mean(-tf.math.log(pos_sim / (pos_sim + Ng)))

    # similarity of positive pairs (only for debug)
    mean_sim = K.mean(tf.linalg.tensor_diag_part(sim))
    mean_neg = K.mean(neg)
    return loss, mean_sim, mean_neg


def fc_loss(
    history,
    future,
    similarity,
    temperature=0.1,
    elimination_topk=0.1,
):  # pragma: no cover
    """Compute the FC loss function, as described in the paper "Fully-Contrastive Learning of Visual Representations" (https://arxiv.org/abs/2004.11362)."""
    N = history.shape[0]
    if elimination_topk > 0.5:
        elimination_topk = 0.5
    elimination_topk = np.math.ceil(elimination_topk * N)

    sim = similarity(history, future) / temperature

    pos_sim = K.exp(tf.linalg.tensor_diag_part(sim))

    tri_mask = np.ones(N**2, dtype=bool).reshape(N, N)
    tri_mask[np.diag_indices(N)] = False
    neg_sim = tf.reshape(tf.boolean_mask(sim, tri_mask), [N, N - 1])

    sorted_sim = tf.sort(neg_sim, axis=1)

    # Top-K cancellation only
    if elimination_topk == 0:
        elimination_topk = 1
    tri_mask = np.ones(N * (N - 1), dtype=bool).reshape(N, N - 1)
    tri_mask[:, -elimination_topk:] = False
    neg = tf.reshape(
        tf.boolean_mask(sorted_sim, tri_mask), [N, N - elimination_topk - 1]
    )
    neg_sim = K.sum(K.exp(neg), axis=1)

    # categorical cross entropy
    loss = K.mean(-tf.math.log(pos_sim / (pos_sim + neg_sim)))

    # similarity of positive pairs (only for debug)
    mean_sim = K.mean(tf.linalg.tensor_diag_part(sim)) * temperature
    mean_neg = K.mean(neg) * temperature
    return loss, mean_sim, mean_neg


def hard_loss(
    history, future, similarity, temperature, beta=0.0, debiased=True, tau_plus=0.1
):  # pragma: no cover
    """Compute the Hard loss function, as described in the paper "Contrastive Learning with Hard Negative Samples" (https://arxiv.org/abs/2011.03343)."""
    N = history.shape[0]

    sim = similarity(history, future)
    pos_sim = K.exp(tf.linalg.tensor_diag_part(sim) / temperature)

    tri_mask = np.ones(N**2, dtype=bool).reshape(N, N)
    tri_mask[np.diag_indices(N)] = False
    neg = tf.reshape(tf.boolean_mask(sim, tri_mask), [N, N - 1])
    neg_sim = K.exp(neg / temperature)

    reweight = (beta * neg_sim) / tf.reshape(tf.reduce_mean(neg_sim, axis=1), [-1, 1])
    if beta == 0:
        reweight = 1
    # estimator g()
    if debiased:
        N = N - 1

        Ng = (-tau_plus * N * pos_sim + tf.reduce_sum(reweight * neg_sim, axis=-1)) / (
            1 - tau_plus
        )
        # constraint (optional)
        Ng = tf.clip_by_value(
            Ng, clip_value_min=np.e ** (-1 / temperature), clip_value_max=tf.float32.max
        )
    else:
        Ng = K.sum(neg_sim, axis=-1)

    # contrastive loss
    loss = K.mean(-tf.math.log(pos_sim / (pos_sim + Ng)))
    # similarity of positive pairs (only for debug)
    mean_sim = K.mean(tf.linalg.tensor_diag_part(sim))
    mean_neg = K.mean(neg)
    return loss, mean_sim, mean_neg


def compute_kmeans_loss(
    latent_means: tf.Tensor, weight: float = 1.0, batch_size: int = 64
):  # pragma: no cover
    """Add a penalty to the singular values of the Gram matrix of the latent means. It helps disentangle the latent space.

    Based on https://arxiv.org/pdf/1610.04794.pdf, and https://www.biorxiv.org/content/10.1101/2020.05.14.095430v3.

    Args:
        latent_means (tf.Tensor): tensor containing the means of the latent distribution
        weight (float): weight of the Gram loss in the total loss function
        batch_size (int): batch size of the data to compute the kmeans loss for.

    Returns:
        tf.Tensor: kmeans loss

    """
    gram_matrix = (tf.transpose(latent_means) @ latent_means) / tf.cast(
        batch_size, tf.float32
    )
    s = tf.linalg.svd(gram_matrix, compute_uv=False)
    s = tf.sqrt(tf.maximum(s, 1e-9))
    return weight * tf.reduce_mean(s)


@tf.function
def get_k_nearest_neighbors(tensor, k, index):  # pragma: no cover
    """Retrieve indices of the k nearest neighbors in tensor to the vector with the specified index.

    Args:
        tensor (tf.Tensor): tensor to compute the k nearest neighbors for
        k (int): number of nearest neighbors to retrieve
        index (int): index of the vector to compute the k nearest neighbors for

    Returns:
        tf.Tensor: indices of the k nearest neighbors

    """
    query = tf.gather(tensor, index, batch_dims=0)
    distances = tf.norm(tensor - query, axis=1)
    max_distance = tf.gather(tf.sort(distances), k)
    neighbourhood_mask = distances < max_distance
    return tf.squeeze(tf.where(neighbourhood_mask))


@tf.function
def compute_shannon_entropy(tensor):  # pragma: no cover
    """Compute Shannon entropy for a given tensor.

    Args:
        tensor (tf.Tensor): tensor to compute the entropy for

    Returns:
        tf.Tensor: entropy of the tensor

    """
    tensor = tf.cast(tensor, tf.dtypes.int32)
    bins = (
        tf.math.bincount(tensor, dtype=tf.dtypes.float32)
        / tf.cast(tf.shape(tensor), tf.dtypes.float32)[0]
    )
    return -tf.reduce_sum(bins * tf.math.log(bins + 1e-5))


def plot_lr_vs_loss(rates, losses):  # pragma: no cover
    """Plot learning rate versus the loss function of the model.

    Args:
        rates (np.ndarray): array containing the learning rates to plot in the x-axis
        losses (np.ndarray): array containing the losses to plot in the y-axis

    """
    plt.plot(rates, losses)
    plt.gca().set_xscale("log")
    plt.hlines(min(losses), min(rates), max(rates))
    plt.axis([min(rates), max(rates), min(losses), (losses[0] + min(losses)) / 2])
    plt.xlabel("Learning rate")
    plt.ylabel("Loss")


def get_angles(pos: int, i: int, d_model: int):
    """Auxiliary function for positional encoding computation.

    Args:
        pos (int): position in the sequence.
        i (int): number of sequences.
        d_model (int): dimensionality of the embeddings.

    """
    angle_rates = 1 / np.power(10000, (2 * (i // 2)) / np.float32(d_model))
    return pos * angle_rates


def get_recurrent_block(
    x: tf.Tensor, latent_dim: int, gru_unroll: bool, bidirectional_merge: str
):
    """Build a recurrent embedding block, using a 1D convolution followed by two bidirectional GRU layers.

    Args:
        x (tf.Tensor): Input tensor.
        latent_dim (int): Number of dimensions of the output tensor.
        gru_unroll (bool): whether to unroll the GRU layers. Defaults to False.
        bidirectional_merge (str): how to merge the forward and backward GRU layers. Defaults to "concat".

    Returns:
        tf.keras.models.Model object with the specified architecture.

    """
    encoder = TimeDistributed(
        tf.keras.layers.Conv1D(
            filters=2 * latent_dim,
            kernel_size=5,
            strides=1,  # Increased strides yield shorter sequences
            padding="same",
            activation="relu",
            kernel_initializer=he_uniform(),
            use_bias=False,
        )
    )(x)
    encoder = tf.keras.layers.Masking(mask_value=0.0)(encoder)
    encoder = TimeDistributed(
        Bidirectional(
            GRU(
                2 * latent_dim,
                activation="tanh",
                recurrent_activation="sigmoid",
                return_sequences=True,
                unroll=gru_unroll,
                use_bias=True,
            ),
            merge_mode=bidirectional_merge,
        )
    )(encoder)
    encoder = LayerNormalization()(encoder)
    encoder = TimeDistributed(
        Bidirectional(
            GRU(
                latent_dim,
                activation="tanh",
                recurrent_activation="sigmoid",
                return_sequences=False,
                unroll=gru_unroll,
                use_bias=True,
            ),
            merge_mode=bidirectional_merge,
        )
    )(encoder)
    encoder = LayerNormalization()(encoder)

    return tf.keras.models.Model(x, encoder)


def positional_encoding(position: int, d_model: int):
    """Compute positional encodings, as in https://proceedings.neurips.cc/paper/2017/file/3f5ee243547dee91fbd053c1c4a845aa-Paper.pdf.

    Args:
        position (int): position in the sequence.
        d_model (int): dimensionality of the embeddings.

    """
    angle_rads = get_angles(
        np.arange(position)[:, np.newaxis], np.arange(d_model)[np.newaxis, :], d_model
    )

    # apply sin to even indices in the array; 2i
    angle_rads[:, 0::2] = np.sin(angle_rads[:, 0::2])

    # apply cos to odd indices in the array; 2i+1
    angle_rads[:, 1::2] = np.cos(angle_rads[:, 1::2])

    pos_encoding = angle_rads[np.newaxis, ...]

    return tf.cast(pos_encoding, dtype=tf.float32)


def create_padding_mask(seq: tf.Tensor):
    """Create a padding mask, with zeros where data is missing, and ones where data is available.

    Args:
        seq (tf.Tensor): Sequence to compute the mask on

    """
    seq = tf.cast(tf.math.equal(seq, 0), tf.float32)

    # add extra dimensions to add the padding to the attention logits.
    return tf.cast(1 - seq[:, tf.newaxis, tf.newaxis, :], tf.float32)


def create_look_ahead_mask(size: int):
    """Create a triangular matrix containing an increasing amount of ones from left to right on each subsequent row.

    Useful for transformer decoder, which allows it to go through the data in a sequential manner, without taking
    the future into account.

    Args:
        size (int): number of time steps in the sequence

    """
    mask = tf.linalg.band_part(tf.ones((size, size)), -1, 0)
    return tf.cast(mask, tf.float32)


def create_masks(inp: tf.Tensor):
    """Given an input sequence, it creates all necessary masks to pass it through the transformer architecture.

    This includes encoder and decoder padding masks, and a look-ahead mask

    Args:
        inp (tf.Tensor): input sequence to create the masks for.

    """
    # Reduces the dimensionality of the mask to remove the feature dimension
    tar = inp[:, :, 0]
    inp = inp[:, :, 0]

    # Encoder padding mask
    enc_padding_mask = create_padding_mask(inp)

    # Used in the 2nd attention block in the decoder.
    # This padding mask is used to mask the encoder outputs.
    dec_padding_mask = create_padding_mask(inp)

    # Used in the 1st attention block in the decoder.
    # It is used to pad and mask future tokens in the input received by
    # the decoder.
    look_ahead_mask = create_look_ahead_mask(tf.shape(tar)[1])
    dec_target_padding_mask = create_padding_mask(tar)
    combined_mask = tf.maximum(dec_target_padding_mask, look_ahead_mask)

    return enc_padding_mask, combined_mask, dec_padding_mask


def find_learning_rate(
    model, data, epochs=1, batch_size=32, min_rate=10**-8, max_rate=10**-1
):
    """Train the provided model for an epoch with an exponentially increasing learning rate.

    Args:
        model (tf.keras.Model): model to train
        data (tuple): training data
        epochs (int): number of epochs to train the model for
        batch_size (int): batch size to use for training
        min_rate (float): minimum learning rate to consider
        max_rate (float): maximum learning rate to consider

    Returns:
        float: learning rate that resulted in the lowest loss

    """
    init_weights = model.get_weights()
    iterations = len(data)
    factor = K.exp(K.log(max_rate / min_rate) / iterations)
    init_lr = K.get_value(model.optimizer.lr)
    K.set_value(model.optimizer.lr, min_rate)
    exp_lr = ExponentialLearningRate(factor)
    model.fit(data, epochs=epochs, batch_size=batch_size, callbacks=[exp_lr])
    K.set_value(model.optimizer.lr, init_lr)
    model.set_weights(init_weights)
    return exp_lr.rates, exp_lr.losses


@tf.function
def get_hard_counts(soft_counts: tf.Tensor):
    """Compute hard counts per cluster in a differentiable way.

    Args:
        soft_counts (tf.Tensor): soft counts per cluster

    """
    max_per_row = tf.expand_dims(tf.reduce_max(soft_counts, axis=1), axis=1)

    mask = tf.cast(soft_counts == max_per_row, tf.float32)
    mask_forward = tf.multiply(soft_counts, mask)
    mask_complement = tf.multiply(1 / soft_counts, mask)

    binary_counts = tf.multiply(mask_forward, mask_complement)

    return tf.math.reduce_sum(binary_counts, axis=0) + 1


@tf.function
def cluster_frequencies_regularizer(
    soft_counts: tf.Tensor, k: int, n_samples: int = 1000
):
    """Compute the KL divergence between the cluster assignment distribution and a uniform prior across clusters.

    While this assumes an equal distribution between clusters, the prior can be tweaked to reflect domain knowledge.

    Args:
        soft_counts (tf.Tensor): soft counts per cluster
        k (int): number of clusters
        n_samples (int): number of samples to draw from the categorical distribution modeling cluster assignments.

    """
    hard_counts = get_hard_counts(soft_counts)

    dist_a = tfd.Categorical(probs=hard_counts / k)
    dist_b = tfd.Categorical(probs=tf.ones(k))

    z = dist_a.sample(n_samples)

    return tf.reduce_mean(dist_a.log_prob(z) - dist_b.log_prob(z))


def get_callbacks(
    embedding_model: str,
    encoder_type: str,
    kmeans_loss: float = 1.0,
    input_type: str = False,
    cp: bool = False,
    logparam: dict = None,
    outpath: str = ".",
    run: int = False,
) -> List[Union[Any]]:
    """Generate callbacks used for model training.

    Args:
        embedding_model (str): name of the embedding model
        encoder_type (str): Architecture used for the encoder. Must be one of "recurrent", "TCN", and "transformer"
        kmeans_loss (float): Weight of the gram loss
        input_type (str): Input type to use for training
        cp (bool): Whether to use checkpointing or not
        logparam (dict): Dictionary containing the hyperparameters to log in tensorboard
        outpath (str): Path to the output directory
        run (int): Run number to use for checkpointing

    Returns:
        List[Union[Any]]: List of callbacks to be used for training

    """
    run_ID = "{}{}{}{}{}{}{}".format(
        "deepof_unsupervised_{}_{}_encodings".format(embedding_model, encoder_type),
        ("_input_type={}".format(input_type if input_type else "coords")),
        ("_kmeans_loss={}".format(kmeans_loss)),
        ("_encoding={}".format(logparam["latent_dim"]) if logparam is not None else ""),
        ("_k={}".format(logparam["n_components"]) if logparam is not None else ""),
        ("_run={}".format(run) if run else ""),
        ("_{}".format(datetime.now().strftime("%Y%m%d-%H%M%S")) if not run else ""),
    )

    log_dir = os.path.abspath(os.path.join(outpath, "fit", run_ID))
    tensorboard_callback = tf.keras.callbacks.TensorBoard(
        log_dir=log_dir, histogram_freq=1, profile_batch=2
    )

    reduce_lr_callback = tf.keras.callbacks.ReduceLROnPlateau(
        monitor="val_loss", factor=0.5, patience=10, cooldown=5, min_lr=1e-8
    )

    callbacks = [run_ID, tensorboard_callback, reduce_lr_callback]

    if cp:
        cp_callback = tf.keras.callbacks.ModelCheckpoint(
            os.path.join(outpath, "checkpoints", run_ID + "/cp-{epoch:04d}.ckpt"),
            verbose=1,
            save_best_only=False,
            save_weights_only=True,
            save_freq="epoch",
        )
        callbacks.append(cp_callback)

    return callbacks


class CustomStopper(tf.keras.callbacks.EarlyStopping):
    """Custom early stopping callback. Prevents the model from stopping before warmup is over."""

    def __init__(self, start_epoch, *args, **kwargs):
        """Initialize the CustomStopper callback.

        Args:
            start_epoch: epoch from which performance will be taken into account when deciding whether to stop training.
            *args: arguments passed to the callback.
            **kwargs: keyword arguments passed to the callback.

        """
        super(CustomStopper, self).__init__(*args, **kwargs)
        self.start_epoch = start_epoch

    def get_config(self):  # pragma: no cover
        """Update callback metadata."""
        config = super().get_config().copy()
        config.update({"start_epoch": self.start_epoch})
        return config

    def on_epoch_end(self, epoch, logs=None):
        """Check whether to stop training."""
        if epoch > self.start_epoch:
            super().on_epoch_end(epoch, logs)


class ExponentialLearningRate(tf.keras.callbacks.Callback):
    """Simple class that allows to grow learning rate exponentially during training.

    Used to trigger optimal learning rate search in deepof.train_utils.find_learning_rate.

    """

    def __init__(self, factor: float):
        """Initialize the exponential learning rate callback.

        Args:
            factor(float): factor by which to multiply the learning rate

        """
        super().__init__()
        self.factor = factor
        self.rates = []
        self.losses = []

    # noinspection PyMethodOverriding
    def on_batch_end(self, batch: int, logs: dict):
        """Apply on batch end.

        Args:
            batch: batch number
            logs (dict): dictionary containing the loss for the current batch

        """
        self.rates.append(K.get_value(self.model.optimizer.lr))
        if "total_loss" in logs.keys():
            self.losses.append(logs["total_loss"])
        else:
            self.losses.append(logs["loss"])
        K.set_value(self.model.optimizer.lr, self.model.optimizer.lr * self.factor)


class ProbabilisticDecoder(tf.keras.layers.Layer):
    """Map the reconstruction output of a given decoder to a multivariate normal distribution."""

    def __init__(self, input_shape, **kwargs):
        """Initialize the probabilistic decoder."""
        super().__init__(**kwargs)
        self.time_distributer = tf.keras.layers.Dense(
            tfpl.IndependentNormal.params_size(input_shape[1:]) // 2
        )
        self.probabilistic_decoding = tfpl.DistributionLambda(
            make_distribution_fn=lambda decoded: tfd.Masked(
                tfd.Independent(
                    tfd.Normal(
                        loc=decoded[0],
                        scale=tf.ones_like(decoded[0]),
                        validate_args=False,
                        allow_nan_stats=False,
                    ),
                    reinterpreted_batch_ndims=1,
                ),
                validity_mask=decoded[1],
            ),
            convert_to_tensor_fn="mean",
        )
        self.scaled_probabilistic_decoding = tfpl.DistributionLambda(
            make_distribution_fn=lambda decoded: tfd.Masked(
                tfd.TransformedDistribution(
                    decoded[0],
                    tfb.Scale(tf.cast(tf.expand_dims(decoded[1], axis=2), tf.float32)),
                    name="vae_reconstruction",
                ),
                validity_mask=decoded[1],
            ),
            convert_to_tensor_fn="mean",
        )

    def call(self, inputs):  # pragma: no cover
        """Map the reconstruction output of a given decoder to a multivariate normal distribution.

        Args:
            inputs (tuple): tuple containing the reconstruction output and the validity mask

        Returns:
            tf.Tensor: multivariate normal distribution

        """
        hidden, validity_mask = inputs

        hidden = tf.keras.layers.TimeDistributed(self.time_distributer)(hidden)
        prob_decoded = self.probabilistic_decoding([hidden, validity_mask])
        scaled_prob_decoded = self.scaled_probabilistic_decoding(
            [prob_decoded, validity_mask]
        )

        return scaled_prob_decoded


class ClusterControl(tf.keras.layers.Layer):
    """Identity layer.

    Evaluates different clustering metrics between the components of the latent Gaussian Mixture
    using the entropy of the nearest neighbourhood. If self.loss_weight > 0, it also adds a regularization
    penalty to the loss function which attempts to maximize the number of populated clusters during training.

    """

    def __init__(
        self,
        batch_size: int,
        n_components: int,
        encoding_dim: int,
        k: int = 15,
        *args,
        **kwargs,
    ):
        """Initialize the ClusterControl layer.

        Args:
            batch_size (int): batch size of the model
            n_components (int): number of components in the latent Gaussian Mixture
            encoding_dim (int): dimension of the latent Gaussian Mixture
            k (int): number of nearest components of the latent Gaussian Mixture to consider
            loss_weight (float): weight of the regularization penalty applied to the local entropy of each training instance
            *args: additional positional arguments
            **kwargs: additional keyword arguments

        """
        super(ClusterControl, self).__init__(*args, **kwargs)
        self.batch_size = batch_size
        self.n_components = n_components
        self.enc = encoding_dim
        self.k = k

    def get_config(self):  # pragma: no cover
        """Update Constraint metadata."""
        config = super().get_config().copy()
        config.update({"batch_size": self.batch_size})
        config.update({"n_components": self.n_components})
        config.update({"enc": self.enc})
        config.update({"k": self.k})
        config.update({"loss_weight": self.loss_weight})
        return config

    def call(self, inputs):  # pragma: no cover
        """Update Layer's call method."""
        encodings, categorical = inputs[0], inputs[1]

        hard_groups = tf.math.argmax(categorical, axis=1)
        max_groups = tf.reduce_max(categorical, axis=1)

        n_components = tf.cast(
            tf.shape(
                tf.unique(tf.reshape(tf.cast(hard_groups, tf.dtypes.float32), [-1]))[0]
            )[0],
            tf.dtypes.float32,
        )

        self.add_metric(n_components, name="number_of_populated_clusters")
        self.add_metric(
            max_groups, aggregation="mean", name="confidence_in_selected_cluster"
        )

        return encodings


class TransformerEncoderLayer(tf.keras.layers.Layer):
    """Transformer encoder layer. Based on https://www.tensorflow.org/text/tutorials/transformer."""

    def __init__(self, key_dim, num_heads, dff, rate=0.1):
        """Construct the transformer encoder layer.

        Args:
            key_dim: dimensionality of the time series
            num_heads: number of heads of the multi-head-attention layers
            dff: dimensionality of the embeddings
            rate: dropout rate

        """
        super(TransformerEncoderLayer, self).__init__()

        self.mha = tf.keras.layers.MultiHeadAttention(
            num_heads=num_heads, key_dim=key_dim
        )
        self.ffn = tf.keras.Sequential(
            [
                tf.keras.layers.Dense(
                    dff, activation="relu"
                ),  # (batch_size, seq_len, dff)
                tf.keras.layers.Dense(key_dim),  # (batch_size, seq_len, d_model)
            ]
        )

        self.layernorm1 = tf.keras.layers.LayerNormalization(epsilon=1e-6)
        self.layernorm2 = tf.keras.layers.LayerNormalization(epsilon=1e-6)

        self.dropout1 = tf.keras.layers.Dropout(rate)
        self.dropout2 = tf.keras.layers.Dropout(rate)

    def call(self, x, training, mask, return_scores=False):  # pragma: no cover
        """Call the transformer encoder layer."""
        attn_output, attn_scores = self.mha(
            key=x, query=x, value=x, attention_mask=mask, return_attention_scores=True
        )  # (batch_size, input_seq_len, d_model)
        attn_output = self.dropout1(attn_output, training=training)
        out1 = self.layernorm1(x + attn_output)

        ffn_output = self.ffn(out1)
        ffn_output = self.dropout2(ffn_output, training=training)
        out2 = self.layernorm2(out1 + ffn_output)

        if return_scores:  # pragma: no cover
            return out2, attn_scores

        return out2


class TransformerDecoderLayer(tf.keras.layers.Layer):
    """Transformer decoder layer. Based on https://www.tensorflow.org/text/tutorials/transformer."""

    def __init__(self, key_dim, num_heads, dff, rate=0.1):
        """Construct the transformer decoder layer.

        Args:
            key_dim: dimensionality of the time series
            num_heads: number of heads of the multi-head-attention layers
            dff: dimensionality of the embeddings
            rate: dropout rate

        """
        super(TransformerDecoderLayer, self).__init__()

        self.mha1 = tf.keras.layers.MultiHeadAttention(
            num_heads=num_heads, key_dim=key_dim
        )
        self.mha2 = tf.keras.layers.MultiHeadAttention(
            num_heads=num_heads, key_dim=key_dim
        )

        self.ffn = tf.keras.Sequential(
            [
                tf.keras.layers.Dense(dff, activation="relu"),
                tf.keras.layers.Dense(key_dim),
            ]
        )

        self.layernorm1 = tf.keras.layers.LayerNormalization(epsilon=1e-6)
        self.layernorm2 = tf.keras.layers.LayerNormalization(epsilon=1e-6)
        self.layernorm3 = tf.keras.layers.LayerNormalization(epsilon=1e-6)

        self.dropout1 = tf.keras.layers.Dropout(rate)
        self.dropout2 = tf.keras.layers.Dropout(rate)
        self.dropout3 = tf.keras.layers.Dropout(rate)

    def call(
        self, x, enc_output, training, look_ahead_mask, padding_mask
    ):  # pragma: no cover
        """Call the transformer decoder layer."""
        attn1, attn_weights_block1 = self.mha1(
            key=x,
            query=x,
            value=x,
            attention_mask=look_ahead_mask,
            return_attention_scores=True,
        )  # (batch_size, target_seq_len, d_model)
        attn1 = self.dropout1(attn1, training=training)
        out1 = self.layernorm1(attn1 + x)

        attn2, attn_weights_block2 = self.mha2(
            key=enc_output,
            query=out1,
            value=enc_output,
            attention_mask=padding_mask,
            return_attention_scores=True,
        )  # (batch_size, target_seq_len, d_model)
        attn2 = self.dropout2(attn2, training=training)
        out2 = self.layernorm2(attn2 + out1)

        ffn_output = self.ffn(out2)
        ffn_output = self.dropout3(ffn_output, training=training)
        out3 = self.layernorm3(ffn_output + out2)

        return out3, attn_weights_block1, attn_weights_block2


# noinspection PyCallingNonCallable
class TransformerEncoder(tf.keras.layers.Layer):
    """Transformer encoder.

    Based on https://www.tensorflow.org/text/tutorials/transformer.
    Adapted according to https://academic.oup.com/gigascience/article/8/11/giz134/5626377?login=true
    and https://arxiv.org/abs/1711.03905.

    """

    def __init__(
        self,
        num_layers,
        seq_dim,
        key_dim,
        num_heads,
        dff,
        maximum_position_encoding,
        rate=0.1,
    ):
        """Construct the transformer encoder.

        Args:
            num_layers: number of transformer layers to include.
            seq_dim: dimensionality of the sequence embeddings
            key_dim: dimensionality of the time series
            num_heads: number of heads of the multi-head-attention layers used on the transformer encoder
            dff: dimensionality of the token embeddings
            maximum_position_encoding: maximum time series length
            rate: dropout rate

        """
        super(TransformerEncoder, self).__init__()

        self.rate = rate
        self.key_dim = key_dim
        self.num_layers = num_layers
        self.embedding = tf.keras.layers.Conv1D(
            key_dim, kernel_size=1, activation="relu", input_shape=[seq_dim, key_dim]
        )
        self.pos_encoding = positional_encoding(maximum_position_encoding, self.key_dim)
        self.enc_layers = [
            TransformerEncoderLayer(key_dim, num_heads, dff, rate)
            for _ in range(num_layers)
        ]
        self.dropout = tf.keras.layers.Dropout(self.rate)

    def call(self, x, training):  # pragma: no cover
        """Call the transformer encoder."""
        # compute mask on the fly
        mask, _, _ = create_masks(x)
        seq_len = tf.shape(x)[1]

        # adding embedding and position encoding.
        x = self.embedding(x)
        x *= tf.math.sqrt(tf.cast(self.key_dim, tf.float32))
        x += self.pos_encoding[:, :seq_len, :]
        x = self.dropout(x, training=training)

        for i in range(self.num_layers):
            x = self.enc_layers[i](x, training, mask)

        return x


# noinspection PyCallingNonCallable
class TransformerDecoder(tf.keras.layers.Layer):
    """Transformer decoder.

    Based on https://www.tensorflow.org/text/tutorials/transformer.
    Adapted according to https://academic.oup.com/gigascience/article/8/11/giz134/5626377?login=true
    and https://arxiv.org/abs/1711.03905.

    """

    def __init__(
        self,
        num_layers,
        seq_dim,
        key_dim,
        num_heads,
        dff,
        maximum_position_encoding,
        rate=0.1,
    ):
        """Construct the transformer decoder.

        Args:
            num_layers: number of transformer layers to include.
            seq_dim: dimensionality of the sequence embeddings
            key_dim: dimensionality of the time series
            num_heads: number of heads of the multi-head-attention layers used on the transformer encoder
            dff: dimensionality of the token embeddings
            maximum_position_encoding: maximum time series length
            rate: dropout rate

        """
        super(TransformerDecoder, self).__init__()

        self.rate = rate
        self.key_dim = key_dim
        self.num_layers = num_layers
        self.embedding = tf.keras.layers.Conv1D(
            key_dim, kernel_size=1, activation="relu", input_shape=[seq_dim, key_dim]
        )
        self.pos_encoding = positional_encoding(maximum_position_encoding, key_dim)
        self.dec_layers = [
            TransformerDecoderLayer(key_dim, num_heads, dff, rate)
            for _ in range(num_layers)
        ]
        self.dropout = tf.keras.layers.Dropout(self.rate)

    def call(
        self, x, enc_output, training, look_ahead_mask, padding_mask
    ):  # pragma: no cover
        """Call the transformer decoder."""
        seq_len = tf.shape(x)[1]
        attention_weights = {}
        x = self.embedding(x)
        x *= tf.math.sqrt(tf.cast(self.key_dim, tf.float32))
        x += self.pos_encoding[:, :seq_len, :]
        x = self.dropout(x, training=training)

        for i in range(self.num_layers):
            x, block1, block2 = self.dec_layers[i](
                x, enc_output, training, look_ahead_mask, padding_mask
            )
            attention_weights["decoder_layer{}_block1".format(i + 1)] = block1
            attention_weights["decoder_layer{}_block2".format(i + 1)] = block2

        return x, attention_weights


def log_hyperparameters():
    """Log hyperparameters in tensorboard.

    Blueprint for hyperparameter and metric logging in tensorboard during hyperparameter tuning

    Returns:
        logparams (list): List containing the hyperparameters to log in tensorboard.
        metrics (list): List containing the metrics to log in tensorboard.

    """
    logparams = [
        hp.HParam(
            "latent_dim",
            hp.Discrete([2, 4, 6, 8, 12, 16]),
            display_name="latent_dim",
            description="encoding size dimensionality",
        ),
        hp.HParam(
            "n_components",
            hp.IntInterval(min_value=1, max_value=25),
            display_name="n_components",
            description="latent component number",
        ),
        hp.HParam(
            "kmeans_weight",
            hp.RealInterval(min_value=0.0, max_value=1.0),
            display_name="kmeans_weight",
            description="weight of the kmeans loss",
        ),
    ]

    metrics = [
        hp.Metric(
            "val_number_of_populated_clusters",
            display_name="number of populated clusters",
        ),
        hp.Metric("val_reconstruction_loss", display_name="reconstruction loss"),
        hp.Metric("val_kmeans_loss", display_name="kmeans loss"),
        hp.Metric("val_vq_loss", display_name="vq loss"),
        hp.Metric("val_total_loss", display_name="total loss"),
    ]

    return logparams, metrics



def embedding_model_fitting(
    preprocessed_object: Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray],
    adjacency_matrix: np.ndarray,
    embedding_model: str,
    encoder_type: str,
    batch_size: int,
    latent_dim: int,
    epochs: int,
    log_history: bool,
    log_hparams: bool,
    n_components: int,
    output_path: str,
    data_path: str,
    kmeans_loss: float,
    pretrained: str,
    save_checkpoints: bool,
    save_weights: bool,
    input_type: str,
    bin_info: dict,
    # VaDE Model specific parameters
    kl_annealing_mode: str,
    kl_warmup: int,
    reg_cat_clusters: float,
    recluster: bool,
    # Contrastive Model specific parameters
    temperature: float,
    contrastive_similarity_function: str,
    contrastive_loss_function: str,
    beta: float,
    tau: float,
    interaction_regularization: float,
    run: int = 0,
    **kwargs,
):
    """

    Trains the specified embedding model on the preprocessed data.

    Args:
        preprocessed_object (tuple): Tuple containing the preprocessed data.
        adjacency_matrix (np.ndarray): adjacency_matrix (np.ndarray): adjacency matrix of the connectivity graph to use.
        embedding_model (str): Model to use to embed and cluster the data. Must be one of VQVAE (default), VaDE, and contrastive.
        encoder_type (str): Encoder architecture to use. Must be one of "recurrent", "TCN", and "transformer".
        batch_size (int): Batch size to use for training.
        latent_dim (int): Encoding size to use for training.
        epochs (int): Number of epochs to train the autoencoder for.
        log_history (bool): Whether to log the history of the autoencoder.
        log_hparams (bool): Whether to log the hyperparameters used for training.
        n_components (int): Number of components to fit to the data.
        output_path (str): Path to the output directory.
        data_path (str): Path to the directory where intermediate data is saved
        kmeans_loss (float): Weight of the gram loss, which adds a regularization term to VQVAE models which penalizes the correlation between the dimensions in the latent space.
        pretrained (str): Path to the pretrained weights to use for the autoencoder.
        save_checkpoints (bool): Whether to save checkpoints during training.
        save_weights (bool): Whether to save the weights of the autoencoder after training.
        input_type (str): Input type of the TableDict objects used for preprocessing. For logging purposes only.
        bin_info (dict): Dictionary containing numpy integer arrays for each experiment. Each array denotes the samples to be sampled from the respective experiment.

        # VaDE Model specific parameters
        kl_annealing_mode (str): Mode to use for KL annealing. Must be one of "linear" (default), or "sigmoid".
        kl_warmup (int): Number of epochs during which KL is annealed.
        reg_cat_clusters (bool): whether to penalize uneven cluster membership in the latent space, by minimizing the KL divergence between cluster membership and a uniform categorical distribution.
        recluster (bool): Whether to recluster the data after each training using a Gaussian Mixture Model.

        # Contrastive Model specific parameters
        temperature (float): temperature parameter for the contrastive loss functions. Higher values put harsher penalties on negative pair similarity.
        contrastive_similarity_function (str): similarity function between positive and negative pairs. Must be one of 'cosine' (default), 'euclidean', 'dot', and 'edit'.
        contrastive_loss_function (str): contrastive loss function. Must be one of 'nce' (default), 'dcl', 'fc', and 'hard_dcl'. See specific documentation for details.
        beta (float): Beta (concentration) parameter for the hard_dcl contrastive loss. Higher values lead to 'harder' negative samples.
        tau (float): Tau parameter for the dcl and hard_dcl contrastive losses, indicating positive class probability.
        interaction_regularization (float): Weight of the interaction regularization term (L1 penalization to all features not related to interactions).
        run (int): Run number to use for logging.



    Returns:
        List of trained models corresponding to the selected model class. The full trained model is last.

    """
    # Select strategy based on available hardware
    if len(tf.config.list_physical_devices("GPU")) > 1:  # pragma: no cover
        strategy = tf.distribute.MirroredStrategy(
            [dev.name for dev in tf.config.list_physical_devices("GPU")]
        )
    elif len(tf.config.list_physical_devices("GPU")) == 1:
        strategy = tf.distribute.OneDeviceStrategy("gpu")
    else:
        strategy = tf.distribute.OneDeviceStrategy("cpu")

    with tf.device("CPU"):

        # Load data
        preprocessed_train, preprocessed_validation= preprocessed_object
        
        X_train, a_train, _ = preprocessed_train.sample_windows_from_data(time_bin_info=bin_info, return_edges=True)
        X_val, a_val, _ = preprocessed_validation.sample_windows_from_data(time_bin_info=bin_info, return_edges=True)


        # Make sure that batch_size is not larger than training set
        if batch_size > X_train.shape[0]:
            batch_size = X_train.shape[0]

        # Set options for tf.data.Datasets
        options = tf.data.Options()
        options.experimental_distribute.auto_shard_policy = (
            tf.data.experimental.AutoShardPolicy.DATA
        )

        # Defines hyperparameters to log on tensorboard (useful for keeping track of different models)
        logparam = {
            "latent_dim": latent_dim,
            "n_components": n_components,
            "kmeans_weight": kmeans_loss,
        }

        # Load callbacks
        run_ID, *cbacks = get_callbacks(
            embedding_model=embedding_model,
            encoder_type=encoder_type,
            kmeans_loss=kmeans_loss,
            input_type=input_type,
            cp=save_checkpoints,
            logparam=logparam,
            outpath=output_path,
            run=run,
        )
        if not log_history:
            cbacks = cbacks[1:]

        Xs, ys = X_train, [X_train]
        Xvals, yvals = X_val, [X_val]
        
        train_shape=X_train.shape
        a_train_shape=a_train.shape


        # Cast to float32
        ys = tuple([tf.cast(dat, tf.float32) for dat in ys])
        yvals = tuple([tf.cast(dat, tf.float32) for dat in yvals])

        # Convert data to tf.data.Dataset objects
        train_dataset = (
            tf.data.Dataset.from_tensor_slices(
                (tf.cast(Xs, tf.float32), tf.cast(a_train, tf.float32), tuple(ys))
            )
            .batch(batch_size * strategy.num_replicas_in_sync, drop_remainder=True)
            .shuffle(buffer_size=train_shape[0])
            .with_options(options)
            .prefetch(tf.data.AUTOTUNE)
        )
        val_dataset = (
            tf.data.Dataset.from_tensor_slices(
                (tf.cast(Xvals, tf.float32), tf.cast(a_val, tf.float32), tuple(yvals))
            )
            .batch(batch_size * strategy.num_replicas_in_sync, drop_remainder=True)
            .with_options(options)
            .prefetch(tf.data.AUTOTUNE)
        )


        embed_x={}
        train_path = os.path.join(data_path, 'embed_x')
        embed_x['embed_x'] = save_dt(Xs,train_path,True)
        embed_a={}
        train_path = os.path.join(data_path, 'embed_a')
        embed_a['embed_a'] = save_dt(a_train,train_path,True) 

        del Xs
        del X_train
        del a_train
        del X_val
        del Xvals
        del a_val
        del ys
        del yvals


    # Build model
    with strategy.scope():

        if embedding_model == "VQVAE":
            ae_full_model = deepof.models.VQVAE(
                input_shape=train_shape,
                edge_feature_shape=a_train_shape,
                adjacency_matrix=adjacency_matrix,
                latent_dim=latent_dim,
                use_gnn=not np.all(np.abs(get_dt(embed_a, 'embed_a')) < 10e-10),
                n_components=n_components,
                kmeans_loss=kmeans_loss,
                encoder_type=encoder_type,
                interaction_regularization=interaction_regularization,
            )
            ae_full_model.optimizer = tf.keras.optimizers.Nadam(
                learning_rate=1e-4, clipvalue=0.75
            )

        elif embedding_model == "VaDE":
            ae_full_model = deepof.models.VaDE(
                input_shape=train_shape,
                edge_feature_shape=a_train_shape,
                adjacency_matrix=adjacency_matrix,
                batch_size=batch_size,
                latent_dim=latent_dim,
                use_gnn=not np.all(np.abs(get_dt(embed_a, 'embed_a')) < 10e-10),
                kl_annealing_mode=kl_annealing_mode,
                kl_warmup_epochs=kl_warmup,
                montecarlo_kl=100,
                n_components=n_components,
                reg_cat_clusters=reg_cat_clusters,
                encoder_type=encoder_type,
                interaction_regularization=interaction_regularization,
            )

        elif embedding_model == "Contrastive":
            ae_full_model = deepof.models.Contrastive(
                input_shape=train_shape,
                edge_feature_shape=a_train_shape,
                adjacency_matrix=adjacency_matrix,
                latent_dim=latent_dim,
                use_gnn=not np.all(np.abs(get_dt(embed_a, 'embed_a')) < 10e-10),
                encoder_type=encoder_type,
                temperature=temperature,
                similarity_function=contrastive_similarity_function,
                loss_function=contrastive_loss_function,
                interaction_regularization=interaction_regularization,
                beta=beta,
                tau=tau,
            )

        else:  # pragma: no cover
            raise ValueError(
                "Invalid embedding model. Select one of 'VQVAE', 'VaDE', and 'Contrastive'"
            )

    callbacks_ = cbacks + [
        CustomStopper(
            monitor="val_total_loss",
            mode="min",
            patience=15,
            restore_best_weights=False,
            start_epoch=15,
        )
    ]

    ae_full_model.compile(
        optimizer=ae_full_model.optimizer,
        run_eagerly=False,
    )

    if not pretrained:
        if embedding_model == "VaDE":
            ae_full_model.pretrain(
                train_dataset,
                embed_x=embed_x,
                embed_a=embed_a,
                epochs=(np.minimum(10, epochs) if not pretrained else 0),
                **kwargs,
            )
            ae_full_model.optimizer._iterations.assign(0)


        ae_full_model.fit(
            x=train_dataset,
            epochs=epochs,
            validation_data=val_dataset,
            callbacks=callbacks_,
            **kwargs,
        )

        if embedding_model == "VaDE" and recluster == True:  # pragma: no cover
            ae_full_model.pretrain(
                train_dataset, embed_x=embed_x, embed_a=embed_a, epochs=0, **kwargs
            )

    else:  # pragma: no cover
        # If pretrained models are specified, load weights and return
        if embedding_model != "Contrastive":
            ae_full_model.build([train_shape, a_train_shape])
        else:
            ae_full_model.build(
                [
                    (train_shape[0], train_shape[1] // 2, train_shape[2]),
                    (a_train_shape[0], a_train_shape[1] // 2, a_train_shape[2]),
                ]
            )

        ae_full_model.load_weights(pretrained)
        return ae_full_model

    if not os.path.exists(os.path.join(output_path, "trained_weights")):
        os.makedirs(os.path.join(output_path, "trained_weights"))

    if save_weights:
        ae_full_model.save_weights(
            os.path.join(
                "{}".format(output_path),
                "trained_weights",
                "{}_final_weights.h5".format(run_ID),
            )
        )

        # Logs hyperparameters to tensorboard
        if log_hparams:
            logparams, metrics = log_hyperparameters()

            tb_writer = tf.summary.create_file_writer(
                os.path.abspath(os.path.join(output_path, "hparams", run_ID))
            )
            with tb_writer.as_default():
                # Configure hyperparameter logging in tensorboard
                hp.hparams_config(hparams=logparams, metrics=metrics)
                hp.hparams(logparam)  # Log hyperparameters

                # Log metrics
                tf.summary.scalar(
                    "val_total_loss",
                    ae_full_model.history.history["val_total_loss"][-1],
                    step=0,
                )

                if embedding_model != "Contrastive":
                    tf.summary.scalar(
                        "val_reconstruction_loss",
                        ae_full_model.history.history["val_reconstruction_loss"][-1],
                        step=0,
                    )
                    tf.summary.scalar(
                        "val_number_of_populated_clusters",
                        ae_full_model.history.history[
                            "val_number_of_populated_clusters"
                        ][-1],
                        step=0,
                    )
                    tf.summary.scalar(
                        "val_kmeans_loss",
                        ae_full_model.history.history["val_kmeans_loss"][-1],
                        step=0,
                    )

                if embedding_model == "VQVAE":
                    tf.summary.scalar(
                        "val_vq_loss",
                        ae_full_model.history.history["val_vq_loss"][-1],
                        step=0,
                    )

                elif embedding_model == "VaDE":
                    tf.summary.scalar(
                        "val_kl_loss",
                        ae_full_model.history.history["val_kl_divergence"][-1],
                        step=0,
                    )

                elif embedding_model == "Contrastive":
                    tf.summary.scalar(
                        "val_total_loss",
                        ae_full_model.history.history["val_total_loss"][-1],
                        step=0,
                    )

    return ae_full_model


def embedding_per_video(
    coordinates: coordinates,
    to_preprocess: table_dict,
    model: tf.keras.models.Model,
    scale: str = "standard",
    animal_id: str = None,
    global_scaler: Any = None,
	pretrained: bool = False,
	samples_max: int = 227272,
    **kwargs,
):  # pragma: no cover
    """Use a previously trained model to produce embeddings and soft_counts per experiment in table_dict format.

    Args:
        coordinates (coordinates): deepof.Coordinates object for the project at hand.
        to_preprocess (table_dict): dictionary with (merged) features to process.
        model (tf.keras.models.Model): trained deepof unsupervised model to run inference with.
        pretrained (bool): whether to use the specified pretrained model to recluster the data.
        scale (str): The type of scaler to use within animals. Defaults to 'standard', but can be changed to 'minmax', 'robust', or False. Use the same that was used when training the original model.
        animal_id (str): if more than one animal is present, provide the ID(s) of the animal(s) to include.
        global_scaler (Any): trained global scaler produced when processing the original dataset.
        samples_max (int): Maximum number of samples taken for plotting to avoid excessive computation times. If the number of rows in a data set exceeds this number the data is downsampled accordingly.
        **kwargs: additional arguments to pass to coordinates.get_graph_dataset().

    Returns:
        embeddings (table_dict): embeddings per experiment.
        soft_counts (table_dict): soft_counts per experiment.

    """

    # at some point _check_enum_inputs will get moved somewhere else and be reworked to function as a general guard function 
    deepof.visuals_utils._check_enum_inputs(
        coordinates,
        animal_id=animal_id,
    )

    embeddings = {}
    soft_counts = {}
    #interim
    file_name='unsup'


    graph = False
    contrastive = isinstance(model, deepof.models.Contrastive)
    try:
        if any([isinstance(i, CensNetConv) for i in model.encoder.layers[2].layers]):
            graph = True
    except AttributeError:
        if any([isinstance(i, CensNetConv) for i in model.encoder.layers]):
            graph = True

    keys_to_drop=[]
    window_size = model.layers[0].input_shape[0][1]
    for key in tqdm.tqdm(to_preprocess.keys(), desc=f"{'Computing embeddings':<{PROGRESS_BAR_FIXED_WIDTH}}", unit="table"):

        dict_to_preprocess = to_preprocess.filter_videos([key])
        #preload datatable in case it is not already, as this will only contain a single table and hence avoid double loading in get_graph_dataset
        dict_to_preprocess[key]=get_dt(dict_to_preprocess,key)
        if dict_to_preprocess[key].isna().all().all():
            keys_to_drop.append(key)
            continue

        #creates a new line to ensure that the outer loading bar does not get overwritten by the inner ones
        print("")

        if graph:
            processed_exp, _, _, _, _ = coordinates.get_graph_dataset(
                animal_id=animal_id,
                precomputed_tab_dict=dict_to_preprocess,
                preprocess=True,
                scale=scale,
                window_size=window_size,
                window_step=1,
                pretrained_scaler=global_scaler,
                samples_max=samples_max,
            )

        else:

            processed_exp, _, _ = dict_to_preprocess.preprocess(
                coordinates=coordinates,
                scale=scale,
                window_size=window_size,
                window_step=1,
                shuffle=False,
                pretrained_scaler=global_scaler,
            )

        tab_tuple=deepof.utils.get_dt(processed_exp[0],key)

        emb = model.encoder([tab_tuple[0], tab_tuple[1]]).numpy()

        if not contrastive:
            sc = model.grouper(
                [tab_tuple[0], tab_tuple[1]]
            ).numpy()
            # save paths for modified tables
            table_path = os.path.join(coordinates._project_path, coordinates._project_name, 'Tables',key, key + '_' + file_name + '_softc')
            soft_counts[key] = deepof.utils.save_dt(sc,table_path,coordinates._very_large_project)

        # save paths for modified tables
        table_path = os.path.join(coordinates._project_path, coordinates._project_name, 'Tables',key, key + '_' + file_name + '_embed')
        embeddings[key] = deepof.utils.save_dt(emb,table_path,coordinates._very_large_project) 

        #to not flood the output with loading bars
        clear_output()

    # Notify user about key removal, if applicable 
    exp_conds=copy.copy(coordinates.get_exp_conditions)
    if len(keys_to_drop) > 0:
        for key in keys_to_drop:
            del exp_conds[key]
        print(
            f'\033[33mInfo! Removed keys {str(keys_to_drop)} As table segments contained only NaNs!\033[0m'
        )

    if contrastive:
        soft_counts = deepof.post_hoc.recluster(
            coordinates, embeddings, pretrained=pretrained, **kwargs
        )
    if isinstance(soft_counts, tuple):
        soft_counts = soft_counts[0]


    table_path=os.path.join(coordinates._project_path, coordinates._project_name, "Tables")
    return (
        deepof.data.TableDict(
            embeddings,
            typ="unsupervised_embedding",
            table_path=table_path, 
            exp_conditions=exp_conds,
        ),
        deepof.data.TableDict(
            soft_counts,
            typ="unsupervised_counts",
            table_path=table_path, 
            exp_conditions=exp_conds,
        ),
    )


def tune_search(
    preprocessed_object: tuple,
    adjacency_matrix: np.ndarray,
    encoding_size: int,
    embedding_model: str,
    hypertun_trials: int,
    hpt_type: str,
    k: int,
    project_name: str,
    callbacks: List,
    batch_size: int = 1024,
    n_epochs: int = 30,
    n_replicas: int = 1,
    outpath: str = "unsupervised_tuner_search",
) -> tuple:
    """Define the search space using keras-tuner and hyperband or bayesian optimization.

    Args:
        preprocessed_object (tf.data.Dataset): Dataset object for training and validation.
        adjacency_matrix (np.ndarray): Adjacency matrix for the graph.
        encoding_size (int): Size of the encoding layer.
        embedding_model (str): Model to use to embed and cluster the data. Must be one of VQVAE (default), VaDE, and Contrastive.
        hypertun_trials (int): Number of hypertuning trials to run.
        hpt_type (str): Type of hypertuning to run. Must be one of "hyperband" or "bayesian".
        k (int): Number of clusters on the latent space.
        project_name (str): Name of the project.
        callbacks (List): List of callbacks to use.
        batch_size (int): Batch size to use.
        n_epochs (int): Maximum number of epochs to train for.
        n_replicas (int): Number of replicas to use.
        outpath (str): Path to save the results.

    Returns:
        best_hparams (dict): Dictionary of the best hyperparameters.
        best_run (str): Name of the best run.

    """

    # extract from Tuple
    preprocessed_train, preprocessed_validation= preprocessed_object
    pt_shape=get_dt(preprocessed_train,list(preprocessed_train.keys())[0], only_metainfo=True)['shape']

    #get available memory -10% as buffer
    available_mem=psutil.virtual_memory().available*0.9
    #calculate maximum number of rows that fit in memory based on table info 
    N_windows_max=int(available_mem/((pt_shape[1]+11)*pt_shape[2]*8))

    # Sample up to N_windows_max windows from processed_train and processed_validation
    N_windows_tab=int(N_windows_max/(len(preprocessed_train)+len(preprocessed_validation)))
    
    X_train, a_train, _ = preprocessed_train.sample_windows_from_data(N_windows_tab=N_windows_tab, return_edges=True)
    X_val, a_val, _ = preprocessed_validation.sample_windows_from_data(N_windows_tab=N_windows_tab, return_edges=True)

    # Make sure that batch_size is not larger than training set
    if batch_size > X_train.shape[0]:
        batch_size = X_train.shape[0]

    # Set options for tf.data.Datasets
    options = tf.data.Options()
    options.experimental_distribute.auto_shard_policy = (
        tf.data.experimental.AutoShardPolicy.DATA
    )

    Xs, ys = X_train, [X_train]
    Xvals, yvals = X_val, [X_val]

    # Cast to float32
    ys = tuple([tf.cast(dat, tf.float32) for dat in ys])
    yvals = tuple([tf.cast(dat, tf.float32) for dat in yvals])

    # Convert data to tf.data.Dataset objects
    train_dataset = (
        tf.data.Dataset.from_tensor_slices(
            (tf.cast(Xs, tf.float32), tf.cast(a_train, tf.float32), tuple(ys))
        )
        .batch(batch_size, drop_remainder=True)
        .shuffle(buffer_size=X_train.shape[0])
        .with_options(options)
        .prefetch(tf.data.AUTOTUNE)
    )
    val_dataset = (
        tf.data.Dataset.from_tensor_slices(
            (tf.cast(Xvals, tf.float32), tf.cast(a_val, tf.float32), tuple(yvals))
        )
        .batch(batch_size, drop_remainder=True)
        .with_options(options)
        .prefetch(tf.data.AUTOTUNE)
    )

    assert hpt_type in ["bayopt", "hyperband"], (
        "Invalid hyperparameter tuning framework. " "Select one of bayopt and hyperband"
    )

    if embedding_model == "VQVAE":
        hypermodel = deepof.hypermodels.VQVAE(
            input_shape=X_train.shape,
            edge_feature_shape=a_train.shape,
            use_gnn=len(preprocessed_object) == 6,
            adjacency_matrix=adjacency_matrix,
            latent_dim=encoding_size,
            n_components=k,
        )
    elif embedding_model == "VaDE":
        hypermodel = deepof.hypermodels.VaDE(
            input_shape=X_train.shape,
            edge_feature_shape=a_train.shape,
            use_gnn=len(preprocessed_object) == 6,
            adjacency_matrix=adjacency_matrix,
            latent_dim=encoding_size,
            n_components=k,
            batch_size=batch_size,
        )
    elif embedding_model == "Contrastive":
        hypermodel = deepof.hypermodels.Contrastive(
            input_shape=X_train.shape,
            edge_feature_shape=a_train.shape,
            use_gnn=len(preprocessed_object) == 6,
            adjacency_matrix=adjacency_matrix,
            latent_dim=encoding_size,
        )

    tuner_objective = "val_total_loss"

    # noinspection PyUnboundLocalVariable
    hpt_params = {
        "hypermodel": hypermodel,
        "executions_per_trial": n_replicas,
        "objective": Objective(tuner_objective, direction="min"),
        "project_name": project_name,
        "tune_new_entries": True,
    }

    if hpt_type == "hyperband":
        tuner = Hyperband(
            directory=os.path.join(
                outpath, "HyperBandx_VQVAE_{}".format(str(date.today()))
            ),
            max_epochs=n_epochs,
            hyperband_iterations=hypertun_trials,
            factor=3,
            **hpt_params,
        )
    else:
        tuner = BayesianOptimization(
            directory=os.path.join(
                outpath, "BayOpt_VQVAE_{}".format(str(date.today()))
            ),
            max_trials=hypertun_trials,
            **hpt_params,
        )

    print(tuner.search_space_summary())

    # Convert data to tf.data.Dataset objects
    tuner.search(
        train_dataset,
        epochs=n_epochs,
        validation_data=val_dataset,
        verbose=1,
        batch_size=batch_size,
        callbacks=callbacks,
    )

    best_hparams = tuner.get_best_hyperparameters(num_trials=1)[0]
    best_run = tuner.hypermodel.build(best_hparams)

    print(tuner.results_summary())

    return best_hparams, best_run
